#include <stdio.h>
#include <time.h>
#include <vector>
#include <iostream>

#include "util.h"
#include "NN.h"
#include "PSO.h"

using namespace std;

/** Usage: ./program  train  <nnet_config_file>  <training_samples_file>  [<PSO_parameters_file> [<output_network_file]]
           ./program  <network_file>
           ./program  timings <nnet_config_file>  <training_samples_file>  [<PSO_parameters_file>]
*/

void print_usage_info(int argc, char *argv[]);

int main(int argc, char *argv[])
{
    srand((unsigned)time(NULL));

    event_pair timer;


    if(argc < 2) print_usage_info(argc,argv);


    if(strcmp("train",argv[1]) == 0 )
    {
        if(argc < 4) print_usage_info(argc,argv);

        string nnetConfigFile = argv[2];
        string trainingDataFile = argv[3];
        string psoConfigFile;
        if(argc >= 5) psoConfigFile = argv[4];

        NeuralNetworks nets(nnetConfigFile, trainingDataFile, psoConfigFile);
        //nets.printInfo();

        start_timer(&timer);
        nets.host_SimulateNeuralNetworks();
        check_cuda_error("C++ host");
        stop_timer(&timer,"C++ host");

        vector<float> host_out = nets.getOutputs();

        start_timer(&timer);
        nets.device_SimulateNeuralNetworks();
        check_cuda_error("C++ device");
        stop_timer(&timer,"C++ device");


        start_timer(&timer);
        vector<float> host_mse = nets.host_computeMSE();
        check_cuda_error("C++ host mse");
        stop_timer(&timer,"C++ host mse");

        start_timer(&timer);
        vector<float> device_mse = nets.device_computeMSE();
        check_cuda_error("C++ device mse");
        stop_timer(&timer,"C++ device mse");

        vector<float> device_out = nets.getOutputs();


        start_timer(&timer);
        nets.train();
        check_cuda_error("NN training");
        stop_timer(&timer,"NN training");

        string networkOutputFileName = "network.nn";
        if(argc>=6) networkOutputFileName = argv[6];
         nets.saveWeights(networkOutputFileName);

    }
    else if(strcmp("timings",argv[1]) == 0)
    {
        if(argc < 4) print_usage_info(argc,argv);

        string nnetConfigFile = argv[2];
        string trainingDataFile = argv[3];
        string psoConfigFile;
        if(argc >= 5) psoConfigFile = argv[4];

        NeuralNetworks nets(nnetConfigFile, trainingDataFile, psoConfigFile);
        nets.printInfo();


        const int nMeasurements = 5;

        nets.psoParameters_.iterations=nMeasurements;

        double CPU_time = 0;
        double GPU_shared_time = 0;
        double GPU_local_time = 0;

        for(int i=0; i<nMeasurements; i++)
        {
            start_timer(&timer);
            nets.host_SimulateNeuralNetworks();
            check_cuda_error("NN CPU simulate networks");
            CPU_time += elapsed_time(&timer);

            start_timer(&timer);
            nets.device_SimulateNeuralNetworks();
            check_cuda_error("NN GPU simulate networks");
            GPU_shared_time += elapsed_time(&timer);

            start_timer(&timer);
            nets.device_SimulateNeuralNetworks_forceLocalMem();
            check_cuda_error("NN GPU simulate networks force local mem");
            GPU_local_time += elapsed_time(&timer);
        }
        CPU_time /= nMeasurements;
        GPU_shared_time /= nMeasurements;
        GPU_local_time /= nMeasurements;
        int nSims = nets.getNumSimulations();

        printf("mem read = %d\n",nets.memSizeReadFromGlobal()/1024);
        printf("CPU time = %.1f ms  (%.2f Mnets/s)\n", CPU_time, nSims/CPU_time/1000);
        printf("GPU time(local) = %.1f ms  (%.2f Mnets/s)\n", GPU_local_time, nSims/GPU_local_time/1000);
        printf("GPU time(shared) = %.1f ms  (%.2f Mnets/s) (%.1fMB/s)\n",
               GPU_shared_time, nSims/GPU_shared_time/1000, nets.memSizeReadFromGlobal() / GPU_shared_time / 1000);
        printf("GPU speedup = %.1fx\n", CPU_time/GPU_shared_time);



        // now with PSO
        double PSO_CPU_time,PSO_GPU_shared_time, PSO_GPU_local_time;
        PSO_CPU_time = nets.train(NeuralNetworks::force_CPU) / nMeasurements;
        PSO_GPU_shared_time = nets.train(NeuralNetworks::force_GPU_shared) / nMeasurements;
        PSO_GPU_local_time = nets.train(NeuralNetworks::force_GPU_local) / nMeasurements;

        printf("PSO CPU time = %.1f ms  (%.2f Miters/s)\n", PSO_CPU_time, nSims/PSO_CPU_time/1000);
        printf("PSO GPU time(local) = %.1f ms  (%.2f Miters/s)\n", PSO_GPU_local_time, nSims/PSO_GPU_local_time/1000);
        printf("PSO GPU time(shared) = %.1f ms  (%.2f Miters/s) (%.1fMB/s)\n",
               PSO_GPU_shared_time, nSims/PSO_GPU_shared_time/1000, nets.memSizeReadFromGlobal() / PSO_GPU_shared_time / 1000);
        printf("PSO GPU speedup = %.1fx\n", PSO_CPU_time/PSO_GPU_shared_time);


        // different number of samples
        FILE *file = fopen("stats.cvs","w");
        fprintf(file,"Population size=%d\n",nets.getNumberOfNetworks());
        fprintf(file,"Number of mesurements=%d\n",nMeasurements);
        fprintf(file,"Number of weights=%d;Number of threads=%d\n",nets.getNumWeights(),nets.getNumSimulations());


        fprintf(file,"Number of samples;CPU time;GPU time(shared);GPU time(local);GPU speedup;"
                    "PSO CPU time;PSO GPU time(shared);PSO GPU time(local);PSO GPU speedup\n");

        int startingNumSamples = nets.getNumSamples();



        for(int i = 100; i>0; i--)
        {

            double CPU_time = 0;
            double GPU_shared_time = 0;
            double GPU_local_time = 0;

            nets.reduceNumSamples(startingNumSamples * i /100);
            printf("i=%d\tNumber of samples=%d\n",i,nets.getNumSamples());

            vector<float> weights = nets.getWeights();

            for(int j=0; j<nMeasurements; j++)
            {
                start_timer(&timer);
                nets.host_SimulateNeuralNetworks();
                check_cuda_error("NN CPU simulate networks");
                CPU_time += elapsed_time(&timer);

                start_timer(&timer);
                nets.updateWeights(weights);
                nets.device_SimulateNeuralNetworks();
                check_cuda_error("NN GPU simulate networks");
                GPU_shared_time += elapsed_time(&timer);

                start_timer(&timer);
                nets.updateWeights(weights);
                nets.device_SimulateNeuralNetworks_forceLocalMem();
                check_cuda_error("NN GPU simulate networks force local mem");
                GPU_local_time += elapsed_time(&timer);
            }
            CPU_time /= nMeasurements;
            GPU_shared_time /= nMeasurements;
            GPU_local_time /= nMeasurements;

            // with PSO (training)
            PSO_CPU_time = nets.train(NeuralNetworks::force_CPU) / nMeasurements;
            PSO_GPU_shared_time = nets.train(NeuralNetworks::force_GPU_shared) / nMeasurements;
            PSO_GPU_local_time = nets.train(NeuralNetworks::force_GPU_local) / nMeasurements;

            fprintf(file,"%d;%.1f;%.1f;%.1f;%.1f;%.1f;%.1f;%.1f;%.1f\n",nets.getNumSamples(),CPU_time,GPU_shared_time,GPU_local_time, CPU_time/GPU_shared_time,
                    PSO_CPU_time,PSO_GPU_shared_time,PSO_GPU_local_time, PSO_CPU_time/PSO_GPU_shared_time);

        }

        fclose(file);

        // training samples are now unusable

    }
    else if(argc == 2)  // get output/simulate network
    {
        NeuralNetworks nets(argv[1]);

        cout << "Neural network loaded.\n\n";
        while(1)
        {
            cout << "Input:";
            vector<float> in(nets.getInputSize());
            for(int i=0;i<nets.getInputSize();i++)
                cin >> in[i];
            cout << "Output: ";
            for(int i=0;i<nets.getOutputSize();i++)
                cout << nets(in)[i] << " ";
            cout << '\n' << endl;
        }

    }
    else print_usage_info(argc,argv);

    return 0;
}

void print_usage_info(int argc, char *argv[])
{
    if(argc < 2) cerr << "Wrong number of command line arguments." << endl;
    cout << "Usage: " << argv[0] << " train  <nnet_config_file>  <training_samples_file>  [<PSO_parameters_file> [<output_network_file]]" << endl;
    cout << "       " << argv[0] << " <network_file>" << endl;
    cout << "       " << argv[0] << " timings <nnet_config_file>  <training_samples_file>  [<PSO_parameters_file>]" << endl;
    exit(-1);
}
