#ifndef NN_H
#define NN_H

/*  Widest layer hard coded(1000) in slow GPU version
    No copy constructor*/

#include "PSO.h"

__device__ __host__ float sigmoid(float x);
__device__ __host__ void propagate_layer(float *input, float *weights, float *output,int previous_layer_size,int layer_size, int isLastLayer=false);
__device__ __host__ void simulate_neural_network(float *input, float *weights, float *output,int num_layers,int *layers, float *ranges, float *temp_output, int lastLayerLinear=false);


template <int widest_layer>
__global__ void global_simulate_neural_networks_local_mem(float *inputs, float *weights, float *outputs, int num_layers,int *layers, int number_of_networks, int number_of_inputs, int num_weights_);
__global__ void global_simulate_neural_networks_shared_mem(float *inputs, float *weights, float *outputs, int num_layers,int *layers, int number_of_networks, int number_of_inputs, int num_weights_, int widest_layer_);
__global__ void device_compute_mse(float *ouputs, float *targets, float *MSEs, int number_of_networks, int num_test_samples, int output_size);

__host__ __device__ int num_weights(int num_layers,int *layers);
__host__ __device__ int widest_layer(int num_layers, int *layers);


std::vector<int> read_in_nnet_config(char *file_name);
void read_in_training_set(char *file_name, std::vector<float> &input, std::vector<float> &output);



/** Contains several neural networks and training data */
class NeuralNetworks
{

public:
    NeuralNetworks(std::string configFile, std::string trainingDataFile, int numberOfNetworks, float lowerWeightLimit = -1, float upperWeightLimit = 1, bool lastLayerLinear = false);
    NeuralNetworks(std::string nnetConfigFile, std::string trainingDataFile, std::string psoConfigFile);
    NeuralNetworks(std::string weightsFile);
    ~NeuralNetworks();
    void updateWeights(const std::vector<float> &weights);
    void saveWeights(std::string fileName);
    std::vector<float>& getWeights();
    void host_SimulateNeuralNetworks();
    void device_SimulateNeuralNetworks();

    void device_SimulateNeuralNetworks_forceLocalMem();

    std::vector<float> operator() (std::vector<float> input);
    std::vector<float>& getOutputs();
    std::vector<float> host_computeMSE(); // mse(host_ouput_vector - host_target_vector) for each network
    std::vector<float> device_computeMSE(); // mse(device_output_vector - device_target_vector) for each network

    /** Train neural networks with PSO until one condition is met.
        If a condition is zero it will not be used.
        Time is in seconds. */
    enum {no_force=0,force_CPU, force_GPU_shared, force_GPU_local};
    double train(int forceSimulationType=0);

    /** Returns maximum difference between h_outputs_ and h_targets_ */
    float maxError();

    void printInfo();
    void GPU_CPU_output_difference(float &average,float &max); // for testing purposes only

    /** Returns number of input neurons. */
    int getInputSize();
    /** Returns number of output neurons. */
    int getOutputSize();
    /** Returns total number of weights in one network. */
    int getNumWeights();
    /** Returns number of layers in a neural network */
    int getNumLayers();
    /** Returns widest layer in a network.
        Used to allocate memory for simulate_neural_network() */
    int getWidestLayer();
    /** Returns number of training samples (input-ouput pairs). */
    int getNumSamples();
    /** Returns number of different weight sets there are. */
    int getNumberOfNetworks();
    /** Returns how many network simulations are preformed */
    int getNumSimulations();

    int getThreadsPerBlock();
    /** Returns number of bytes read from global memory by all the threads. */
    int memSizeReadFromGlobal();
    int getNumBlocks();

    /** Just for timing purposes. Reduces number of samples to num.*/
    void reduceNumSamples(int num);

//private:
public:

    PsoParameters psoParameters_;

private:

    //for setting up global function parameters
    int blockSize_;
    bool useSharedMemoryInSimulation_;
    bool lastLayerLinear_;
    int maxBlockSize(); // for available shared memory
    int sharedMemoryRequirements(int blockSize, bool includingArguments = false); //shared memory per block for this network

    void readInNNetConfig(std::string fileName);

    /* File must have number of inputs and number of ouputs in the first row;
        then ranges for each ouput (eg. for 2 outputs in range [0,2]: "0 2 0 2");
        then the inputs and outputs, each sample in diffrenet row. */
    void readInTrainingData(std::string fileName);

    std::vector<float> scaleInputData(std::vector<float> in);

    std::vector<float> h_weights_;
    std::vector<float> h_inputs_;
    std::vector<float> h_targets_; //expected outputs
    std::vector<float> h_outputs_;
    std::vector<float> in_ranges_; // inputs are scaled when read in (from file, and from operator() )
    std::vector<float> h_out_ranges_; // [[0],[1]] - range for first output; [[2],[3]] - 2nd output...
    std::vector<int> h_layers_; //number of neurons in each layer [i,h1,..,hn,o]



    // device pointers
    float *d_weights_;
    float *d_inputs_;
    float *d_targets_;
    float *d_outputs_; //final result of all the networks
    float *d_out_ranges_;
    int *d_layers_;

    float *d_MSEs_; // not cached (yet), must calculate every time

    // just for CPU nnet simulation, do no use
    float *h_output_;
    float *h_temp_output_;
};



#endif // NN_H
